/*
 * File:   BaudRate.c
 * Author: Bob Alexander
 *
 * Created on November 24, 2014, 9:46 PM
 */

#include "Common.h"
#include "app.h"
#include "BaudRate.h"
#include "DisplayManager.h"
#include "SerialDataTransfer.h"

#define AUTOBAUD (-1)

static size_t baud;
static const size_t choices[] = {300, 1200, 2400, 4800, 9600, 19200, 38400, 57600, 115200, AUTOBAUD};
static uint32_t lastButtonPressedTime = 0, buttonPressedTime = 0;
static bool baudButtonClicked = false;
static bool saveBaud = false;
static bool autobaudOn = false;
static const size_t flashBlock[DRV_NVM_PAGE_SIZE / sizeof(size_t)] __attribute__((aligned(DRV_NVM_PAGE_SIZE))) = {9600};

struct Autobaud
{
    uint32_t lastEdgeTime;
    uint32_t fastestBitTime;
    int edgeCount;
} autobaud;
static void StartAutobaud();
static void StopAutobaud();
static void DoAutobaud();

static const int buttonTimeout = 3000;

void BaudRate_Initialize()
{
    // Retrieve the last baud rate we were set to. Don't just read the flashBlock array because
    // the optimizer will compile that as the load of a constant
    DRV_HANDLE nvm = DRV_NVM_Open(0, DRV_IO_INTENT_READ);
    DRV_NVM_Read(nvm, (uint8_t *) &baud, (uint8_t *) flashBlock, sizeof(baud));
    DRV_NVM_Close(nvm);

    /* Enable CN14 which is the baud rate button */
    CNENBSET = (1 << _CNENB_CNIEB12_POSITION);

    /* Enable the change notice interrupt source, set its priority level to 2,
       set its subpriority level to 0 */
    PLIB_INT_SourceEnable(INT_ID_0, INT_SOURCE_CHANGE_NOTICE_B);
    PLIB_INT_VectorPrioritySet(INT_ID_0, INT_VECTOR_CN, INT_PRIORITY_LEVEL2);
    PLIB_INT_VectorSubPrioritySet(INT_ID_0, INT_VECTOR_CN, INT_SUBPRIORITY_LEVEL0);
    // There might just have been a spurious Change Notification, so ignore any
    // perceived button click
    baudButtonClicked = false;

    lastButtonPressedTime = -buttonTimeout;

    DisplayManager_ShowBaudRate(baud, buttonTimeout);
}

void BaudRate_Tasks()
{
    if (baudButtonClicked)
    {
        baudButtonClicked = false;

        if (autobaudOn)
        {
            StopAutobaud();
            baud = 300;
            saveBaud = true;
            DisplayManager_ShowBaudRate(baud, buttonTimeout);
        }

        else
        {
            // The first time the button is pressed, the display turns on with the
            // current baud rate. But we check here to see if the button is being
            // pressed repeatedly, in which case we select the next baud rate.
            if (buttonPressedTime - lastButtonPressedTime < buttonTimeout)
            {
                // Find the next baud rate
                int i;
                for (i = 0; i < countof(choices) && choices[i] < baud; ++i)
                {
                }
                if (baud == choices[i])
                {
                    ++i;
                    if (i >= countof(choices))
                        i = 0;
                }
                baud = choices[i];
                if (baud != AUTOBAUD)
                {
                    SerialDataTransfer_SetBaud(baud);
                    saveBaud = true;
                }
                else
                {
                    StartAutobaud();
                    saveBaud = false;
                }
            }
            DisplayManager_ShowBaudRate(baud, baud == DISPLAY_AUTO ? DISPLAY_TIMEOUT_INFINITE : buttonTimeout);
        }
    }

    if (autobaudOn)
    {
        DoAutobaud();
    }

    // If we need to save the baud rate and the button hasn't been pressed for 3 seconds
    if (saveBaud && SYS_TMR_TickCountGet() - buttonPressedTime > 3000)
    {
        size_t flashRow[DRV_NVM_ROW_SIZE / sizeof(size_t)];
        saveBaud = false;
        DRV_HANDLE nvm = DRV_NVM_Open(0, DRV_IO_INTENT_READWRITE);
        DRV_NVM_BUFFER_HANDLE erase = DRV_NVM_Erase(nvm, (uint8_t *) flashBlock, sizeof(flashBlock));
        if (erase != DRV_NVM_BUFFER_HANDLE_INVALID)
        {
            DRV_NVM_BufferStatus(nvm, erase);
            flashRow[0] = baud;
            DRV_NVM_Write(nvm, (uint8_t *) flashBlock, (uint8_t *) flashRow, DRV_NVM_ROW_SIZE);
        }
        DRV_NVM_Close(nvm);
    }
}

/**
 * Interrupt when the baud rate button is pressed
 */
void __attribute__((interrupt(ipl2), vector(_CHANGE_NOTICE_VECTOR))) ChangeNoticeHandler()
{

    /* Read port to clear mismatch on change notice pins */
    int baudRateButtonState = PLIB_PORTS_Read(PORTS_ID_0, PORT_CHANNEL_B) & (1 << 14);

    /* Clear the interrupt flag */
    PLIB_INT_SourceFlagClear(INT_ID_0, INT_SOURCE_CHANGE_NOTICE_B);

    if (baudRateButtonState)
    {
        uint32_t now = SYS_TMR_TickCountGet();
        // If this isn't a key bounce
        if (now - buttonPressedTime > 250)
        {
            lastButtonPressedTime = buttonPressedTime;
            buttonPressedTime = now;
            baudButtonClicked = true;
        }
    }
}

void StartAutobaud()
{
    autobaud.edgeCount = 0;
    autobaud.fastestBitTime = -1;
    autobaud.lastEdgeTime = -1;

    PLIB_TMR_Start(TMR_ID_2);
    PLIB_PORTS_RemapInput(PORTS_ID_0, INPUT_FUNC_IC3, INPUT_PIN_RPB8 );
    DRV_IC0_Enable();
    autobaudOn = true;
}

void StopAutobaud()
{
    DRV_IC0_Disable();
    PLIB_PORTS_RemapInput(PORTS_ID_0, INPUT_FUNC_U2RX, INPUT_PIN_RPB8 );
    PLIB_TMR_Stop(TMR_ID_2);
    autobaudOn = false;
}

void DoAutobaud()
{
    while (!DRV_IC0_BufferIsEmpty())
    {
        uint32_t edgeTime = DRV_IC0_Capture32BitDataRead();
        if (autobaud.edgeCount)
        {
            uint32_t bitTime = edgeTime - autobaud.lastEdgeTime;
            if (bitTime < autobaud.fastestBitTime)
                autobaud.fastestBitTime = bitTime;
        }
        autobaud.lastEdgeTime = edgeTime;
        if (++autobaud.edgeCount > 25)
        {
            StopAutobaud();
            
            uint32_t clockFreq = SYS_CLK_PeripheralFrequencyGet(CLK_BUS_PERIPHERAL_1);
            if (autobaud.fastestBitTime != 0)
            {
                baud = clockFreq / autobaud.fastestBitTime;
                SerialDataTransfer_SetBaud(baud);
                saveBaud = true;
                DisplayManager_ShowBaudRate(baud, buttonTimeout);
            }
            else
            {
                baud = *(size_t *) flashBlock;
                SerialDataTransfer_SetBaud(baud);
                DisplayManager_ShowBaudRate(DISPLAY_ERR, buttonTimeout);
            }
        }
    }
}
